unit Graph2D;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, StdCtrls, ExtCtrls, Math, Controls, Windows, Forms, Graphics,
  PolynomUtil, Settings, dialogs;

type
 TGraph2D = class(TPanel)
  const
   MinRange = 0.000000001;
   MaxRange = 100000000000000;
   PixDiffPerLine = 50;
   Speed = 10;
   physUnitSymbols : Array [0..18] of WideChar =   (  'y',   'z',   'a',   'f',   'p',  'n',  #181,  'u',  'm', #0, 'K', 'k', 'M', 'G',  'T',  'P',  'E',  'Z',  'Y');
   physUnitValues : Array [0..18] of Precision = (1E-24, 1E-21, 1E-18, 1E-15, 1E-12, 1E-9, 1E-6, 1E-6, 1E-3,  1, 1E3, 1E3, 1E6, 1E9, 1E12, 1E15, 1E18, 1E21, 1E24);
  type
   TAxisType = (HORIZONTAL, VERTICAL);
   TLineStyle = (SOLID, DASHED, DOTTED);

   TFloatPoint = packed record
    X : Precision;
    Y : Precision;
   end;

   TAxis = class
    strict private
     Graph : TGraph2D;
    private
     FktMin, FktMax : Precision;
     MoveStartMin   : Precision;
     procedure Render;
    public
     Min, Max      : Precision;
     Color         : TColor;
     Position      : Precision;
     AxisType      : TAxisType;
     Name          : ShortString;
     constructor Create (_AxisType : TAxisType; _Graph : TGraph2D);
     function PosToScreen(v : Precision) : integer;
     function ScreenToPos(v : integer) : Precision;
     procedure SetFKTExtrema;
   end;



   TRenderObj = class
    protected
     Graph : TGraph2D;
     procedure Render; virtual; abstract;
    public
     XAxis, YAxis : TAxis;
     visible : Boolean;
     constructor Create(_Graph : TGraph2D);
     destructor Destroy; override;
   end;



   TFunctionBuff = Array of TFloatPoint;
   TLine = class(TRenderObj)
    private
     procedure Render; override;
    public
     isFunction : Boolean;
     Buff : TFunctionBuff;
     Style : TLineStyle;
     constructor Create (_Graph : TGraph2D);
     destructor Destroy; override;
   end;



   TFunction = class(TRenderObj)
    private
     Buff : TFunctionBuff;
     procedure Render; override;
    public
     Fkt : Poly.IFunction;
     Name : String;
     Style : TLineStyle;
     constructor Create (_Graph : TGraph2D; _fkt : Poly.IFunction);
     destructor Destroy; override;
   end;



   TBar = class(TRenderObj)
    protected
     procedure Render; override;
    public
     BarType : TAxisType;
     Style : TLineStyle;
     Position : Precision;
     constructor Create(_Graph : TGraph2D; _BarType : TAxisType);
   end;



  { TMovePoint = class(TButton)
    private
     _X,_Y      : Double;
     Move       : Boolean;
     StartC     : TPoint;
     StartB     : TPoint;
     procedure MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); reintroduce;
     procedure MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); reintroduce;
     procedure MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer); reintroduce;
    public
     OnMove : TNotifyEvent;
     constructor Create(_Owner : TGraph2D; X,Y : Double);
     property X : Double read _X write _X;
     property Y : Double read _Y write _Y;
     procedure Free;
   end;   }



  strict private
   KeyR    : Boolean;
   KeyL    : Boolean;
   KeyT    : Boolean;
   KeyB    : Boolean;
   ShiftS  : TShiftState;
   Timer   : TTimer;
   LinePos : TPoint;
   LineLength : cardinal;
   _canvas : TCanvas;
   BMP     : TBitmap;

   Move    : Boolean;
   StartC  : TPoint;
   StartPX : Double;
   StartPY : Double;

   KeyCatcher    : TEdit;
   RenderObjects : Array of TRenderObj;
   procedure MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); reintroduce;
   procedure MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); reintroduce;
   procedure MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer); reintroduce;
   procedure MouseWheel(Sender: TObject; Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
   procedure TimerEV(Sender : TObject);
   procedure KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState); reintroduce;
   procedure KeyUp(Sender: TObject; var Key: Word; Shift: TShiftState); reintroduce;
   function readCanvas : TCanvas; inline;
  private
   completeRepaint : Boolean;
   procedure MoveTo(x, y : integer);
   procedure StyledLineTo(x, y : integer; Style : TLineStyle);
   class function IntPower(Basis : Double; Exp : Integer) : Double;
   class procedure GetPosOfRegularPoints(From,UpTo : Double; Count : Word; var Start,Space : Double);
   class function Round(f : precision):integer;
  public
   XAxis : Array of TAxis;
   YAxis : Array of TAxis;
   property Canvas : TCanvas read readCanvas;
   constructor Create(_Owner : TComponent);
   procedure DoPaint(Sender : TObject);
   procedure SaveDataToStream(str : TStream; Resolution : Cardinal = 100);
   class function FloatToPhysStr(f : Double; maxDigits : byte = 10) : String;
   destructor Destroy; override;
 end;

implementation



constructor TGraph2D.TRenderObj.Create(_Graph : TGraph2D);
begin
 inherited Create;
 Graph := _Graph;
 SetLength(Graph.RenderObjects, Length(Graph.RenderObjects)+1);
 Graph.RenderObjects[Length(Graph.RenderObjects)-1] := self;
 XAxis := Graph.XAxis[0];
 YAxis := Graph.YAxis[0];
 visible := true;
end;

destructor TGraph2D.TRenderObj.Destroy;
var
 i : integer;
 b : Boolean;
begin
 b := false;
 for i := 0 to Length(Graph.RenderObjects)-1 do begin
  if (Graph.RenderObjects[i] = self) then b := true;
  if (b) and (i < Length(Graph.RenderObjects)-1) then Graph.RenderObjects[i] := Graph.RenderObjects[i+1];
 end;
 if (b) then SetLength(Graph.RenderObjects, Length(Graph.RenderObjects)-1);
 inherited;
end;





constructor TGraph2D.TLine.Create(_Graph : TGraph2D);
begin
 inherited;
 isFunction := false;
 Style := SOLID;
 Buff := nil;
end;

procedure TGraph2D.TLine.Render;
var i : integer;
begin
 Graph.Canvas.Pen.Color := YAxis.Color;
 if (Length(Buff) < 2) then exit;
 for i := 0 to Length(Buff) - 1 do begin
  if (isFunction) and (Buff[i].x >= XAxis.Min) and (Buff[i].x <= XAxis.Max) then begin
   if (Buff[i].y > YAxis.FktMax) then YAxis.FktMax := Buff[i].y;
   if (Buff[i].y < YAxis.FktMin) then YAxis.FktMin := Buff[i].y;
  end;
  if (i = 0) then Graph.MoveTo(XAxis.PosToScreen(Buff[0].x), YAxis.PosToScreen(Buff[0].y))
  else Graph.StyledLineTo(XAxis.PosToScreen(Buff[i].x), YAxis.PosToScreen(Buff[i].y), Style);
 end;
end;

destructor TGraph2D.TLine.Destroy;
begin
 SetLength(Buff, 0);
 inherited;
end;



constructor TGraph2D.TFunction.Create (_Graph : TGraph2D; _fkt : Poly.IFunction);
begin
 inherited Create(_Graph);
 fkt := _fkt;
 Style := SOLID;
 Name := '';
 Buff := nil;
end;

procedure TGraph2D.TFunction.Render;
var
  b      : Boolean;
  x, i   : integer;
  f      : Precision;
begin
 if (Fkt = nil) then exit;
 Graph.Canvas.Pen.Color := YAxis.Color;
 b := true;
 i := Length(Buff);
 if (i <> Graph.Width) then begin
  SetLength(Buff, Graph.Width);
  for i := i to Graph.Width-1 do Buff[i].x := Poly.NaN;
 end;
 for x := 0 to Graph.Width - 1 do begin
  f := XAxis.ScreenToPos(x);
  if (IsNaN(Buff[x].x)) or (Buff[x].x <> f) or (Graph.completeRepaint) then begin
   Buff[x].x := f;
   Buff[x].y := Fkt.Calc(f);
  end;
  if (Poly.IsNaN(Buff[x].y)) then begin
   b := true;
  end else begin
   if (Buff[x].y < YAxis.FktMin) then YAxis.FktMin := Buff[x].y;
   if (Buff[x].y > YAxis.FktMax) then YAxis.FktMax := Buff[x].y;
   if (b) then begin Graph.MoveTo(x, YAxis.PosToScreen(Buff[x].y)); b := false; end
   else Graph.StyledLineTo(x, YAxis.PosToScreen(Buff[x].y), Style);
  end;
 end;
end;

destructor TGraph2D.TFunction.Destroy;
begin
 fkt.Free;
 inherited;
end;




constructor TGraph2D.TBar.Create(_Graph : TGraph2D; _BarType : TAxisType);
begin
 inherited Create(_Graph);
 BarType := _BarType;
 Position := 0;
 Style := SOLID
end;

procedure TGraph2D.TBar.Render;
var p : integer;
begin
 if (BarType = VERTICAL) then begin
  p := XAxis.PosToScreen(Position);
  Graph.MoveTo(p, 0);
  Graph.StyledLineTo(p, Graph.Height, Style);
 end else begin
  p := XAxis.PosToScreen(Position);
  Graph.MoveTo(0, p);
  Graph.StyledLineTo(Graph.Width, p, Style);
 end;
end;


{constructor TGraph.TMovePoint.Create(_Owner : TGraph2D; X,Y : Double);
begin
 if (not Assigned(_Owner)) then exit;
 inherited Create(_Owner);
 Caption := '';
 Width := 15;
 Height := 15;
 Left := _Owner.XPosToScreen(X) - Width div 2;
 Top := _Owner.YPosToScreen(Y) - Height div 2;
 _X := X;
 _Y := Y;
 Move := false;
 OnMove := nil;
 Parent := _Owner;
 SetLength(_Owner.MovePoints,Length(_Owner.MovePoints) + 1);
 _Owner.MovePoints[Length(_Owner.MovePoints) - 1] := self;
 OnMouseDown := @MouseDown;
 OnMouseUp := @MouseUp;
 OnMouseMove := @MouseMove;
end;

procedure TGraph.TMovePoint.MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 GetCursorPos(StartC);
 StartB.X := Left;
 StartB.Y := Top;
 Move := true;
end;

procedure TGraph.TMovePoint.MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 Move := false;
end;

procedure TGraph.TMovePoint.MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var
 CurPos : TPoint;
begin
 if (Move) then begin
  GetCursorPos(CurPos);
  X := StartB.X + CurPos.X - StartC.X;
  Y := StartB.Y + CurPos.Y - StartC.Y;
  _X := TGraph2D(Owner).XScreenToPos(X);
  _Y := TGraph2D(Owner).YScreenToPos(Y);
  if (OnMove = nil) then begin Left := X; Top := Y; end
  else OnMove(self);
  TGraph2D(Owner).DoPaint(Sender);
 end;
end;

procedure TGraph.TMovePoint.Free;
var i : integer;
begin
 if (Length(TGraph2D(Owner).MovePoints) > 0) then begin
  for i := 0 to Length(TGraph2D(Owner).MovePoints) - 1 do if (TGraph2D(Owner).MovePoints[i] = self) then break;
  if (TGraph2D(Owner).MovePoints[i] = self) then begin
   for i := i to Length(TGraph2D(Owner).MovePoints) - 2 do TGraph2D(Owner).MovePoints[i] := TGraph2D(Owner).MovePoints[i + 1];
   SetLength(TGraph2D(Owner).MovePoints,Length(TGraph2D(Owner).MovePoints) - 1);
  end;
 end;
 inherited Free;
end;  }




constructor TGraph2D.TAxis.Create(_AxisType : TAxisType; _Graph : TGraph2D);
begin
 inherited Create;
 AxisType := _AxisType;
 Graph := _Graph;
 Name := '';
 Color := clBlack;
 Position := 0;
 Min := -1;
 Max := 1;
 if (AxisType = VERTICAL) then begin
  SetLength(Graph.YAxis, Length(Graph.YAxis)+1);
  Graph.YAxis[Length(Graph.YAxis)-1] := self;
 end else begin
  SetLength(Graph.XAxis, Length(Graph.XAxis)+1);
  Graph.XAxis[Length(Graph.XAxis)-1] := self;
 end;
end;

procedure TGraph2D.TAxis.SetFKTExtrema;
begin
 if (FktMin >= FktMax) then exit;
 Min := FktMin - (FktMax - FktMin)/10;
 Max := FktMax + (FktMax - FktMin)/10;
end;

function TGraph2D.TAxis.PosToScreen(v : Precision) : integer;
var s : Precision;
begin
 s := (v - Min) / (Max - Min);  // [Min , Max] auf [0, 1]
 if (AxisType = VERTICAL) then Result := Round((1-s) * Graph.Height) // [0, 1] auf [Height, 0]
 else Result := Round(s * Graph.Width); // [0, 1] auf [0, Width]
end;

function TGraph2D.TAxis.ScreenToPos(v : integer) : Precision;
var s : Precision;
begin
 if (Graph.Width = 0) or (Graph.Height = 0) then begin Result := 0; exit; end;
 if (AxisType = VERTICAL) then s := 1 - v / Graph.Height // [0, Height] auf [1, 0]
 else s := v / Graph.Width; // [0, Width] auf [0, 1]
 Result := Min + s * (Max-Min); // [0, 1] auf [Min, Max]
end;

procedure TGraph2D.TAxis.Render;
var
 Pos : integer;
 LineLength : integer;
 i,j,k      : integer;
 d,e        : double;
 s          : String;
 b          : Byte;
begin
 FktMin := 1E24;
 FktMax := -1E24;
 Graph.Canvas.Pen.Color := Color;
 Graph.Canvas.Brush.Color := clInfoBk;
 Graph.Canvas.Font.Color := Color;
 if (AxisType = VERTICAL) then begin
  Pos := Graph.XAxis[0].PosToScreen(Position);
  if (Pos < 0) then Pos := 0;
  if (Pos >= Graph.Width) then Pos := Graph.Width - 1;
  k := Pos + 1;
  if (k + 20 > Graph.Width) then begin b := 1; k := Graph.Width - 1; end else b := 0;
  LineLength := Floor(Graph.Height / PixDiffPerLine);
  Graph.GetPosOfRegularPoints(Min, Max, LineLength, d, e);
  for i := 0 to LineLength - 1 do begin
   j := PosToScreen((d + e*i));
   if (abs(d + e*i) < e/2) then s := '0'
   else s := Graph.FloatToPhysStr(d + e*i);
   Graph.Canvas.TextOut(k-b*10*Length(s), j - 1, s);
   Graph.Canvas.Line(k - 2,j,k + 2,j);
  end;
  Graph.Canvas.Line(k, 0, k, Graph.Height);
  if (Name <> '') then Graph.Canvas.TextOut(k - b*Length(Name)*10, 0, Name);
 end else begin
  Pos := Graph.YAxis[0].PosToScreen(Position);
  if (Pos < 0) then Pos := 0;
  if (Pos >= Graph.Height) then Pos := Graph.Height - 1;
  k := Pos + 1;
  if (k + 18 > Graph.Height) then k := Graph.Height - 18;
  LineLength := Floor(Graph.Width / PixDiffPerLine);
  Graph.GetPosOfRegularPoints(Min, Max, LineLength, d, e);
  for i := 0 to LineLength - 1 do begin
   j := PosToScreen((d + e*i));
   if (abs(d + e*i) < e/2) then s := '0'
   else s := FloatToPhysStr(d + e*i);
   Graph.Canvas.TextOut(j + 1, k, s);
   Graph.Canvas.Line(j,Pos + 2,j,Pos - 2);
  end;
  Graph.Canvas.Line(0, Pos, Graph.Width, Pos);
 end;
end;




constructor TGraph2D.Create(_Owner : TComponent);
begin
 inherited Create(_Owner);
 LineLength := 0;
 BMP := TBitmap.Create;
 BMP.Width := Width;
 BMP.Height := Height;
 _Canvas := BMP.canvas; //TPanel(self).Canvas;
 RenderObjects := nil;
 completeRepaint := false;
 Self.OnPaint := @DoPaint;
 Move := false;
 OnMouseDown := @MouseDown;
 OnMouseUp := @MouseUp;
 OnMouseMove := @MouseMove;
 OnMouseWheel := @MouseWheel;
 KeyL := false;
 KeyR := false;
 KeyT := false;
 KeyB := false;
 ShiftS := [];
 Timer := TTimer.Create(self);
 Timer.Interval := 40;
 Timer.Enabled := false;
 Timer.OnTimer := @TimerEv;
 OnMouseDown := @MouseDown;
 OnMouseUp := @MouseUp;
 OnMouseMove := @MouseMove;
 OnMouseWheel := @MouseWheel;
 OnKeyDown := @KeyDown;
 OnKeyUp := @KeyUp;
 KeyCatcher := TEdit.Create(self);
 KeyCatcher.Parent := self;
 KeyCatcher.OnKeyDown := @KeyDown;
 KeyCatcher.OnKeyUp := @KeyUp;
 KeyCatcher.ReadOnly := true;
 KeyCatcher.OnEnter := @DoPaint;
 KeyCatcher.OnExit := @DoPaint;
 KeyCatcher.Top := -100;
 XAxis := nil;
 YAxis := nil;
 TAxis.Create(HORIZONTAL, self);
 TAxis.Create(VERTICAL, self)
end;

function TGraph2D.readCanvas : TCanvas; inline;
begin
 Result := _Canvas;
end;

procedure TGraph2D.MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var i : integer;
begin
 KeyCatcher.SetFocus;
 StartC.X := X;
 StartC.Y := Y;
 for i := 0 to Length(XAxis)-1 do XAxis[i].MoveStartMin := XAxis[i].Min;
 for i := 0 to Length(YAxis)-1 do YAxis[i].MoveStartMin := YAxis[i].Min;
 Move := true;
end;

procedure TGraph2D.MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 Move := false;
end;

procedure TGraph2D.MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var
 d : Double;
 i : integer;
begin
 if (Move) then begin
  for i := 0 to Length(XAxis)-1 do begin
   d := XAxis[i].Max - XAxis[i].Min;
   XAxis[i].Min := XAxis[i].MoveStartMin + (StartC.X - X) / Width * d;
   XAxis[i].Max := XAxis[i].Min + d;
  end;
  for i := 0 to Length(YAxis)-1 do begin
   d := YAxis[i].Max - YAxis[i].Min;
   YAxis[i].Min := YAxis[i].MoveStartMin - (StartC.Y - Y) / Height * d;
   YAxis[i].Max := YAxis[i].Min + d;
  end;
  DoPaint(Sender);
 end;
end;

procedure TGraph2D.MouseWheel(Sender: TObject; Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
var
 Sx, Sy, mi, ma : Precision;
 i : integer;
 WheelDeltaX,WheelDeltaY : integer;
begin
 if (Width <= 0) then exit;
 if (Height <= 0) then exit;
 WheelDeltaX := Round(WheelDelta / 1000 * Width);
 WheelDeltaY := Round(WheelDelta / 1000 * Height);
 Sx := MousePos.X / Width;
 Sy := MousePos.Y / Height;
 for i := 0 to Length(XAxis)-1 do begin
  mi := XAxis[i].ScreenToPos(Round(WheelDeltaX * Sx));
  ma := XAxis[i].ScreenToPos(Width - Round(WheelDeltaX * (1 - Sx)));
  XAxis[i].Min := mi;
  XAxis[i].Max := ma;
 end;
 for i := 0 to Length(YAxis)-1 do begin
  ma := YAxis[i].ScreenToPos(Round(WheelDeltaY * Sy));
  mi := YAxis[i].ScreenToPos(Height - Round(WheelDeltaY * (1 - Sy)));
  YAxis[i].Min := mi;
  YAxis[i].Max := ma;
 end;
 DoPaint(Sender);
end;

procedure TGraph2D.TimerEV(Sender : TObject);
var
 SX1,SY1,SX2,SY2 : shortint;
 mi, ma : Precision;
 i : integer;
begin
 if (not KeyT) and (not KeyB) and (not KeyL) and (not KeyR) then begin Timer.Enabled := false; exit; end;
 SX1 := 0;
 SY1 := 0;
 SX2 := 0;
 SY2 := 0;
 if (KeyR) and (not KeyL) then begin SX1 := Speed; SX2 := Speed; end;
 if (KeyL) and (not KeyR) then begin SX1 := -Speed; SX2 := -Speed; end;
 if (KeyR) and (KeyL) and (ssCtrl in ShiftS) then begin SX1 := -Speed; SX2 := Speed; end;
 if (KeyR) and (KeyL) and (not (ssCtrl in ShiftS)) then begin SX1 := Speed; SX2 := -Speed; end;

 if (KeyB) and (not KeyT) then begin SY1 := Speed; SY2 := Speed; end;
 if (KeyT) and (not KeyB) then begin SY1 := -Speed; SY2 := -Speed; end;
 if (KeyT) and (KeyB) and (ssCtrl in ShiftS) then begin SY1 := -Speed; SY2 := Speed; end;
 if (KeyT) and (KeyB) and (not (ssCtrl in ShiftS)) then begin SY1 := Speed; SY2 := -Speed; end;

 for i := 0 to Length(XAxis)-1 do begin
  mi := XAxis[i].ScreenToPos(SX1);
  ma := XAxis[i].ScreenToPos(Width + SX2);
  XAxis[i].Min := mi;
  XAxis[i].Max := ma;
 end;
 for i := 0 to Length(YAxis)-1 do begin
  ma := YAxis[i].ScreenToPos(SY1);
  mi := YAxis[i].ScreenToPos(Height + SY2);
  YAxis[i].Min := mi;
  YAxis[i].Max := ma;
 end;
 DoPaint(Sender);
end;

procedure TGraph2D.KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
 if (Key = 37) then KeyL := true
 else if (Key = 38) then KeyT := true
 else if (Key = 39) then KeyR := true
 else if (Key = 40) then KeyB := true;
 ShiftS := Shift;
 Timer.Enabled := true;
end;

procedure TGraph2D.KeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
 if (Key = 37) then KeyL := false
 else if (Key = 38) then KeyT := false
 else if (Key = 39) then KeyR := false
 else if (Key = 40) then KeyB := false;
 ShiftS := Shift;
end;

procedure TGraph2D.DoPaint(Sender : TObject);
var i : integer;
begin
 if (Sender = self) then completeRepaint := false else completeRepaint := true;
 if (Width <> BMP.Width) or (Height <> BMP.Height) then begin
  BMP.Width := self.Width;
  BMP.Height := self.Height;
 end;
 Canvas.Brush.Color := clInfoBk;
 Canvas.FillRect(0,0,Width,Height);
 for i := 0 to Length(XAxis)-1 do XAxis[i].Render;
 for i := 0 to Length(YAxis)-1 do YAxis[i].Render;
 for i := 0 to Length(RenderObjects)-1 do if (RenderObjects[i].visible) then RenderObjects[i].Render;
 if (KeyCatcher.Focused) then begin
  Canvas.Brush.Color := clBlue;
  Canvas.FrameRect(0,0,Width,Height);
 end;
 TPanel(self).Canvas.Draw(0, 0, BMP);
end;

procedure TGraph2D.MoveTo(x, y : integer);
begin
 LinePos.X := x;
 LinePos.Y := y;
end;

procedure TGraph2D.StyledLineTo(x, y : integer; Style : TLineStyle);
var
 lastPos : TPoint;
 _x, _y, w, m : integer;
 d, e : ShortInt;
begin
 lastPos := LinePos;
 LinePos.X := x;
 LinePos.Y := y;
 if (lastPos.x < 0) and (x < 0) or (lastPos.x > Canvas.Width) and (x > Canvas.Width) or (lastPos.y < 0) and (y < 0) or (lastPos.y > Canvas.Height) and (y > Canvas.Height) then exit; // Linien, die komplett außerhalb liehen, verwerfen
 w := LineLength;
 if (Abs(x-lastPos.x) >= Abs(y-lastPos.y)) then begin
  _x := lastPos.x;
  m := x;
  if (lastPos.x < x) then begin
   d := 1;
   if (_x < 0) then _x := 0;
   if (m > Canvas.Width) then m := Canvas.Width;
  end else begin
   d := -1;
   if (_x > Canvas.Width) then _x := Canvas.Width;
   if (m < 0) then m := 0;
  end;
  if (lastPos.y < y) then e := 1 else e := -1;
  while (_x <> m) do begin
   _y := Trunc(lastPos.y + (Y-lastPos.y+e) / (x-lastPos.x+d) * (_x-lastPos.x));
   case (Style) of
    SOLID : Canvas.Pixels[_x, _y] := Canvas.Pen.Color;
    DASHED : if ((w div 3) mod 2 = 0) then Canvas.Pixels[_x, _y] := Canvas.Pen.Color;
    DOTTED : if (w mod 2 = 0) then Canvas.Pixels[_x, _y] := Canvas.Pen.Color;
   end;
   w := w + 1;
   _x := _x + d;
  end;
 end else begin
  _y := lastPos.y;
  m := y;
  if (lastPos.y < y) then begin
   d := 1;
   if (_y < 0) then _y := 0;
   if (m > Canvas.Height) then m := Canvas.Height;
  end else begin
   d := -1;
   if (_y > Canvas.Height) then _y := Canvas.Height;
   if (m < 0) then m := 0;
  end;
  if (lastPos.x < x) then e := 1 else e := -1;
  while (_y <> m) do begin
   _x := Trunc(lastPos.x + (x-lastPos.x+e) / (y-lastPos.y+d) * (_y-lastPos.y));
   case (Style) of
    SOLID : Canvas.Pixels[_x, _y] := Canvas.Pen.Color;
    DASHED : if ((w div 3) mod 2 = 0) then Canvas.Pixels[_x, _y] := Canvas.Pen.Color;
    DOTTED : if (w mod 2 = 0) then Canvas.Pixels[_x, _y] := Canvas.Pen.Color;
   end;
   w := w + 1;
   _y := _y + d;
  end;
 end;
 LineLength := w;
 if (LineLength > 65535) then LineLength := 0;
 //Canvas.Line(lastPos.x, lastPos.y, x, y);
end;

procedure TGraph2D.SaveDataToStream(str : TStream; Resolution : Cardinal = 100);
var
 fkts : Array of Array of TFunction;
 i,j,k : integer;
 x,y : Precision;
 s : AnsiString;
begin
 if (Length(XAxis) <> 1) then raise Exception.Create('Export for more than one X-Axis not implemented!');
 fkts := nil;
 // alle Funktionen(-Namen) finden
 for i := 0 to Length(RenderObjects)-1 do if (RenderObjects[i] is TFunction) then begin
  if (TFunction(RenderObjects[i]).Name = '') then j := Length(fkts) else j := 0;
  while (j < Length(fkts)) do if (fkts[j][0].Name = TFunction(RenderObjects[i]).Name) then break else j := j + 1;
  if (j = Length(fkts)) then begin SetLength(fkts, j+1); fkts[j] := nil; end;
  k := Length(fkts[j]);
  SetLength(fkts[j], k+1);
  fkts[j][k] := TFunction(RenderObjects[i]);
 end;
 // Kopfzeile schreiben
 s := XAxis[0].Name;
 for i := 0 to Length(fkts)-1 do s := s + #9 + fkts[i][0].Name;
 s := s + #13#10;
 str.Write(s[1], Length(s));
 // Daten schreiben
 for i := 0 to Resolution do begin
  x := XAxis[0].Min + (XAxis[0].Max - XAxis[0].Min) * i / Resolution;
  s := FloatToStr(x);
  for j := 0 to Length(fkts)-1 do begin
   y := Poly.NaN;
   for k := 0 to Length(fkts[j])-1 do begin y := fkts[j][k].Fkt.Calc(x); if (not Poly.IsNaN(y)) then break; end;
   s := s + #9 + FloatToStr(y);
  end;
  s := s + #13#10;
  str.Write(s[1], Length(s));
 end;
end;

destructor TGraph2D.Destroy;
var i : integer;
begin
 for i := Length(RenderObjects)-1 downto 0 do RenderObjects[i].Destroy;
 for i := Length(XAxis)-1 downto 0 do XAxis[i].Destroy;
 for i := Length(YAxis)-1 downto 0 do YAxis[i].Destroy;
 BMP.Free;
 inherited Destroy;
end;





class function TGraph2D.Round(f : precision):integer;
begin
 if (f > MAXINT) then Result := MAXINT
 else if (f < -MAXINT) then Result := - MAXINT
 else Result := System.round(f);
end;

class function TGraph2D.FloatToPhysStr(f : Double; maxDigits : Byte = 10) : String;
var
 g : Double;
 i : integer;
begin
 if (isNaN(f)) then begin Result := 'NaN'; exit; end;
 if (f = 0) then begin Result := '0'; exit; end;
 for i := 0 to Length(physUnitValues)-1 do begin
  g := f / physUnitValues[i];
  if (abs(g) < 1000) then begin
   Result := FloatToStrF(g, FFgeneral, maxDigits, maxDigits);
   if (physUnitSymbols[i] <> #0) then Result := Result + physUnitSymbols[i];
   exit;
  end;
 end;
 Result := FloatToStr(f);
end;

class function TGraph2D.IntPower(Basis : Double; Exp : Integer) : Double;
var i : integer;
begin
 Result := 1;
 for i := 0 to abs(Exp) - 1 do Result := Result * Basis;
 if (Exp < 0) then Result := 1 / Result;
end;

class procedure TGraph2D.GetPosOfRegularPoints(From, UpTo : Double; Count : Word; var Start, Space : Double);
var
 Exp : integer;
 K   : Byte;
 D   : Double;
function GetNextSpace : Double;
begin
 if (K = 5) then K := 2
 else if (K = 2) then K := 1
 else begin
  K := 5;
  dec(Exp);
 end;
 Result := K * IntPower(10,Exp);
end;
begin
 Exp := floor(Ln(Abs(UpTo - From)) / Ln(10)) + 1;
 K := 2;
 Space := K * IntPower(10,Exp);
 repeat begin
  D := GetNextSpace;
  if (D * Count >= UpTo - From) then begin
   Space := D;
   Start := ceil(From / Space) * Space;
  end;
 end until (D * Count < UpTo - From);
 // showmessage('From: ' + floattostr(From) + ' To: ' + floattostr(UpTo) + ' Count: ' + inttostr(Count) + ' Start: ' + floattostr(Start) + ' Space: ' + floattostr(Space));
end;

end.

