unit circuitFrame;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, ExtCtrls, StdCtrls, settings,
  functions, Graph2D, Graphics, ComCtrls, Math, dialogs;

type
  TRelPos = record
   t, l : integer;
   a  : byte;
   el : TControl;
  end;

  { TFrameCircuit }

  TFrameCircuit = class(TFrame)
    CheckBox1: TCheckBox;
    CheckBox2: TCheckBox;
    CheckBox3: TCheckBox;
    CheckBox4: TCheckBox;
    EditIout: TEdit;
    EditCq1: TEdit;
    EditCq2: TEdit;
    EditVin: TEdit;
    EditCres: TEdit;
    EditLs1: TEdit;
    EditLs2: TEdit;
    EditLh: TEdit;
    EditR2: TEdit;
    EditVout: TEdit;
    Image1: TImage;
    Label1: TLabel;
    Label10: TLabel;
    Label11: TLabel;
    Label16: TLabel;
    Label17: TLabel;
    Label18: TLabel;
    LabelAlert: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    RadioButton1: TRadioButton;
    RadioButton2: TRadioButton;
    Timer1: TTimer;
    TrackBar1: TTrackBar;
    procedure CheckBoxChange(Sender: TObject);
    procedure FrameResize(Sender: TObject);
    procedure ParamChange(Sender: TObject);
    procedure SetValues(Sender: TObject);
  private
   relativePositions : Array of TRelPos;
   startH, startW : integer;
  public
    constructor Create(_Owner : TComponent); override;
  end;

implementation

{$R *.lfm}

constructor TFrameCircuit.Create(_Owner : TComponent);
procedure addRel (el : TControl; a : byte);
var n : integer;
begin
 n := Length(relativePositions);
 SetLength(relativePositions, n+1);
 relativePositions[n].el := el;
 relativePositions[n].l := el.Left;
 relativePositions[n].t := el.Top;
 relativePositions[n].a := a
end;
begin
 inherited;
 startH := self.Height;
 startW := self.Width;
 addRel(Label1, 0);
 addRel(Label16, 0);
 addRel(EditCres, 0);
 addRel(EditCq1, 1);
 addRel(EditCq2, 1);
 addRel(EditLs1, 1);
 addRel(EditLs2, 1);
 addRel(EditLh, 0);
 addRel(EditR2, 1);
 addRel(Label9, 1);
end;

procedure TFrameCircuit.FrameResize(Sender: TObject);
var i : integer;
begin
 for i := 0 to Length(relativePositions)-1 do begin
  if (relativePositions[i].a and 1 = 1) then relativePositions[i].el.Top := Trunc((relativePositions[i].t + relativePositions[i].el.Height) / startH * Height) - relativePositions[i].el.Height
  else relativePositions[i].el.Top := Trunc(relativePositions[i].t / startH * Height);
  relativePositions[i].el.Left := Trunc(relativePositions[i].l / startW * Width);
 end;
end;

procedure TFrameCircuit.CheckBoxChange(Sender: TObject);
begin
 if (CheckBox1.Checked) then begin EditLs2.Color := clSilver; EditLs2.Readonly := true; end
 else begin EditLs2.Color := clWhite; EditLs2.Readonly := false; end;
 if (CheckBox2.Checked) then begin EditCres.Color := clSilver; EditCres.Readonly := true; Label16.Visible := true; Trackbar1.Visible := true; end
 else begin EditCres.Color := clWhite; EditCres.Readonly := false; Label16.Visible := false; Trackbar1.Visible := false; end;
 if (CheckBox3.Checked) then begin EditCq2.Color := clSilver; EditCq2.Readonly := true; end
 else begin EditCq2.Color := clWhite; EditCq2.Readonly := false; end;
 if (RadioButton1.Checked) then begin EditIOut.Readonly := false; EditIOut.Color := clWhite; EditVout.ReadOnly := true; EditVOut.Color := clSilver; end
 else begin EditIOut.Readonly := true; EditIOut.Color := clSilver; EditVout.ReadOnly := false; EditVOut.Color := clWhite; end;
 SetValues(Sender);
end;

procedure TFrameCircuit.ParamChange(Sender: TObject);
begin
 Timer1.Enabled := false;
 Timer1.Enabled := true;
end;

procedure TFrameCircuit.SetValues(Sender: TObject);
var
 err : Boolean;
 i : integer;
 T : Precision;
function ValToFloat(El : TEdit) : Precision;
var
 i : integer;
 e : precision;
 v : String;
begin
 e := 1;
 v := El.Text;
 if (Length(v) > 0) then for i := 0 to Length(TGraph2D.physUnitSymbols)-1 do if (TGraph2D.physUnitSymbols[i] = v[Length(v)]) then begin
  e := TGraph2D.physUnitValues[i];
  SetLength(v, Length(v)-1);
  break;
 end;
 try
  Result := StrToFloat(v) * e;
  El.Color := $00FFFFFF;
 except
  Result := Poly.NaN;
  err := true;
  El.Color := $006060FF;
 end;
end;
begin
 Timer1.Enabled := false;
 err := false;
 TSRCValues.Vin := ValToFloat(EditVin);
 if (not RadioButton1.Checked) then TSRCValues.Vout := ValToFloat(EditVout);
 TSRCValues.Cq1 := ValToFloat(EditCq1);
 if (CheckBox3.Checked) then begin TSRCValues.Cq2 := TSRCValues.Cq1; EditCq2.Text := EditCq1.Text; end
 else TSRCValues.Cq2 := ValToFloat(EditCq2);
 TSRCValues.Ls1 := ValToFloat(EditLs1);
 if (CheckBox1.Checked) then begin TSRCValues.Ls2 := TSRCValues.Ls1; EditLs2.Text := EditLs1.Text; end
 else TSRCValues.Ls2 := ValToFloat(EditLs2);
 TSRCValues.Ton := ValToFloat(EditTon);
 TSRCValues.Toff := ValToFloat(EditToff);
 TSRCValues.Lh := ValToFloat(EditLh);
 if (CheckBox2.Checked) then begin
  if (Poly.IsNaN(TSRCValues.Ton)) or (Poly.IsNaN(TSRCValues.Ton)) or (Poly.IsNaN(TSRCValues.Ls1)) or (Poly.IsNaN(TSRCValues.Ls2)) then begin
   TSRCValues.Cres := Poly.NaN;
   EditCres.Text := '';
   Label15.Caption := '';
  end else begin
   TSRCValues.Cres := (2*(TSRCValues.Ton+TSRCValues.Toff))**2 * (Trackbar1.Position * 1.5 / 100 + 0.5)**2 / (4*PI**2*(TSRCValues.Ls1 + TSRCValues.Ls2));
   EditCres.Text := TGraph2D.FloatToPhysStr(TSRCValues.Cres, 4);
   Label15.Caption := TGraph2D.FloatToPhysStr(1/(2*(TSRCValues.Ton+TSRCValues.Toff))) + 'Hz';
  end;
 end else TSRCValues.Cres := ValToFloat(EditCres);
 TSRCValues.R2 := ValToFloat(EditR2);
 TSRCValues.Ih_0_def := 0;
 TSRCValues.Vcres_0_def := 190;
 if (RadioButton1.Checked) then EditVout.Text := TGraph2D.FloatToPhysStr(TSRCValues.gotoSteadyState(ValToFloat(EditIOut) * 2*(TSRCValues.Ton + TSRCValues.Toff)), 4)
 else TSRCValues.update;
 Timer1.Enabled := false;
 LabelAlert.Visible := TSRCValues.isInitVal and not err;
 if (CheckBox4.Checked) and (not Poly.IsNaN(TSRCValues.Ton)) and (not Poly.IsNaN(TSRCValues.Toff)) then begin
  TGraph2D(Graph).XAxis[0].Min := 0;
  TGraph2D(Graph).XAxis[0].Max := 2*(TSRCValues.Ton+TSRCValues.Toff);
  TGraph2D(Graph).doPaint(Sender);
  for i := 0 to Length(TGraph2D(Graph).YAxis)-1 do TGraph2D(Graph).YAxis[i].SetFKTExtrema;
  TGraph2D(Graph).doPaint(Sender);
 end else TGraph2D(Graph).doPaint(Sender);
end;

end.

